// pgsort - narzedzie do sortowania plikow zawierajacych lancuchy znakow (przez strcmp) 

#define USAGE \
	"pgsort [-s rozmiar] [-S pamiec] [-u] [-m pref start stop] [-o wynik] [plik]\n       [--help]\n"

#define HELPTEXT \
"Narzedzie do sortowania plikow zawierajacych lancuchy znakow (przez strcmp)\n\
     -s rozmiar    : maksymalny rozmiar lancucha (domyslnie 100)\n\
     -S pamiec     : ilosc pamieci w megabajtach na jeden blok (domyslnie 768)\n\
     -u            : usuwanie powtorzen\n\
-m pref start stop : laczenie posortowanych plikow pref_start..pref_stop\n\
     -o wynik      : plik wynikowy (domyslnie stdout)\n\
      plik         : plik wejsciowy (domyslnie stdin)\n"

#define SIZE 100
#define MEMORY 768
#define REST 9

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

char ** buffer;
FILE ** files;
char ** first;
int * indeks;
int nr, unique;
FILE *input;
FILE *output;


void go_up(int i){
	int j;
	while (i) {
		j = (i-1)/2;
		
		if (strcmp(first[indeks[i]], first[indeks[j]]) < 0) { // i -ty mniejszy - zamiana i dalej w gore
			int tmp = indeks[i];
			indeks[i] = indeks[j];
			indeks[j] = tmp;
			i = j;
		}
		else break;
	}
}

void go_down(int i) {
	int j = 2*i+1;

	while (j < nr) {
		if (j+1 < nr && strcmp(first[indeks[j]], first[indeks[j+1]]) > 0) j++; // j+1 mniejszy - zmien j
		
		// j - element do porownania
		if (strcmp(first[indeks[i]], first[indeks[j]]) > 0) { // i -ty wiekszy - zamiana i dalej w dol
			int tmp = indeks[i];
			indeks[i] = indeks[j];
			indeks[j] = tmp;
			i = j;
			j = 2*i+1;
		}
		else break;
	}
}

void quicksort(long d, long g) {
	// sortuje buffer[d..g-1]
	if (g - d > REST) {
		long i, s;
		char *t, *t2;
		
		// zamiana pierwszego i rodkowego (nie losowego!)
		i = (d+g)/2;
		t = buffer[i];
		buffer[i] = buffer[d];
		buffer[d] = t;

		// jaowy przebieg - znalec element nie mniejszy ni t
		s = d+1;
		while(s < g && strcmp(buffer[s], t) < 0) s++;
		
		// po pomniejszeniu s o jeden buffer[d+1..s] < t
		// przegldamy reszt
		for (i = s--; i < g; i++)
			if (strcmp(buffer[i], t) < 0) {
				// powikszenie s i zamiana
				t2 = buffer[++s];
				buffer[s] = buffer[i];
				buffer[i] = t2;
			}
		// zamiana buffer[d] z buffer[s]
		buffer[d] = buffer[s];
		buffer[s] = t;
		// teraz buffer[d..s-1] < buffer[s] < buffer[s+1..g-1]
			
		quicksort(d, s); //sortuj [d..s-1]
		quicksort(s+1, g); //sortuj [s+1..g-1]
	}
}

void insertionsort(long d, long g) {
	long i, j;
	char *t;

	for (i = d+1; i < g; i++) {
		t = buffer[i];
		j = i;
		while (j > 0 && strcmp(t, buffer[j-1]) < 0) {
			buffer[j] = buffer[j-1];
			j--;
		}
		buffer[j] = t;
	}
}

int main(int argc, char * argv[])
{
	char prefiks[100], fname[100];
	long start = 0;
	long stop = 0;
	long i; 
	long size = SIZE;
	long mem = MEMORY;
	char *arg;
	int tmp = 0;
	char *last = new char[size];

	unique = 0;
	input = stdin;
	output = stdout;

	if (argc == 1) {
		printf(USAGE);
		return 0;
	}
	else for (i = 1; i < argc; i++) {
		arg = argv[i];
		if (*arg == '-') switch (*(++arg)) {
			case 's' : sscanf(argv[++i], "%d", &size); size++; break;
			case 'S' : sscanf(argv[++i], "%d", &mem); break;
			case 'u' : unique = 1; break;
			case 'm' : strcpy(prefiks, argv[++i]);
                       start = atol(argv[++i]);
					   stop = atol(argv[++i]);
					   break;
			case 'o' : output = fopen(argv[++i], "wt"); break;
			case '-' : printf(USAGE); printf(HELPTEXT); return 0; 
			default  : printf(USAGE); return 0;
		}
		else input = fopen(argv[i], "rt");
	}

	if (!start && !stop) { // sortowanie
	    // rozmiar
		int tmp_size = 8 * ((size+7)/8) + sizeof(buffer);
		long buff_size = 1024; 
		long count;
		FILE * tmp_file;
		buff_size = (buff_size * 1024 * mem)/tmp_size;
		buff_size = 8 * (buff_size/8);

		buffer = new char * [buff_size];

		count = 0;
		while (!feof(input)) {
			if (!tmp) buffer[count]= new char[size];
			fscanf(input, "%s\n", buffer[count++]);
			if (ferror(input)) {
				fprintf(stderr, "Blad odczytu!\n");
				return -1;
			}
        
			if (count == buff_size) {
				// sortowanie skadane
				quicksort(0,buff_size);
				insertionsort(0, buff_size);

				if (!tmp) {
					struct stat b;
					// potrzeba plikw tymczasowych
					tmp = start = stop = 1;
					i = count = 0;
					tmp_file = NULL;
					strcpy(prefiks, "pgsAAAA");
					// proteza - inicjowanie generatora liczb pseudolosowych
					srand(((unsigned)time(NULL))+((unsigned)clock()));
					do {
						// prefiks, fname
						for (i = 3; i <7; i++)	
							prefiks[i] = (char)(rand() % 26) + 'A';
						sprintf(fname, "%s_%d", prefiks, stop);
						// jak sie nie uda 512 razy to zakoncz bledem
						if (count++ == 511) return 1;
						// jeli nie ma pliku o naziw fname to go otwrz do zapisu
						if (stat(fname, &b)) tmp_file = fopen(fname, "wt");
					} while(!tmp_file);
					fprintf(stderr, "pgsort: %s\n", prefiks);
				}
				else {
					sprintf(fname, "%s_%d", prefiks, ++stop);
					tmp_file = fopen(fname, "wt");
				}
				if (unique) {
					last[0] = 0;
					for (i = 0; i < buff_size; i++)
						if (strcmp(last,buffer[i])) { // jesli taki juz byl, to pominac
							strcpy(last, buffer[i]);
							fprintf(tmp_file, "%s\n", last);
							if (ferror(tmp_file)) {
								fprintf(stderr, "Blad!\n");
								return -1;
							}
						}
				}
				else for (i = 0; i < buff_size; i++) {
						fprintf(tmp_file, "%s\n", buffer[i]);
						if (ferror(tmp_file)) {
							fprintf(stderr, "Blad!\n");
							return -1;
						}
				}
				fclose(tmp_file);
				count = 0;
			}
		}
		if (count) { // niepusty bufor - sortowa
			// sortowanie skadane
			quicksort(0,count);
			insertionsort(0, count);
			
			if (tmp) { // wyjcie do pliku tymczasowego lub na output
				sprintf(fname, "%s_%d", prefiks, ++stop);
				tmp_file = fopen(fname, "wt");
			}
			else tmp_file = output;
			if (unique) {
				last[0] = 0;
				for (i = 0; i < count; i++)
					if (strcmp(last,buffer[i])) { // jesli taki juz byl, to pominac
						strcpy(last, buffer[i]);
						fprintf(tmp_file, "%s\n", last);
						if (ferror(tmp_file)) {
							fprintf(stderr, "Blad!\n");
							return -1;
						}
					}
			}
			else for (i = 0; i < count; i++) {
					fprintf(tmp_file, "%s\n", buffer[i]);
					if (ferror(tmp_file)) {
						fprintf(stderr, "Blad!\n");
						return -1;
					}
			}
			if (tmp) fclose(tmp_file);
		}
		
		// sprztanie
		if (start) count = buff_size;
		for (i = count-1; i >= 0; i--)
			delete[] buffer[i];
		delete[] buffer;
	}

	if (start || stop) { // czenie
		files = new FILE *[stop-start+1];	
		first = new char *[stop-start+1];
		indeks = new int[stop-start+1];
	
		for (i = 0; i <= stop-start; i++) {
			first[i] = new char[size]; 	
			sprintf(fname, "%s_%d", prefiks, i+start);
			files[i] = fopen(fname, "rt");
		}

		// sortowanie	
		// wczytac po elemencie z pliku
		nr = 0;
		for (i = 0; i <= stop-start; i++) {
			if (fscanf(files[i], "%s\n", first[i]) == 1) {
				indeks[nr++] = i;
				go_up(nr-1);
			}
			else if (ferror(files[i])) {
				fprintf(stderr, "Blad odczytu!\n");
				return -1;
			}
		}

		// dopoki cos jest wyrzucac element najmniejszy
		if (unique) {
			last[0] = 0;
			while (nr > 0) {
				i = indeks[0];
				if (strcmp(last,first[i])) { // jesli taki juz byl, to pominac
					strcpy(last, first[i]);
					fprintf(output, "%s\n", last);
					if (ferror(output)) {
						fprintf(stderr, "Blad!\n");
						return -1;
					}
				}
				if (fscanf(files[i], "%s\n", first[i]) != 1)	{
					if (ferror(files[i])) {
						fprintf(stderr, "Blad odczytu!\n");
						return -1;
					}
				   else indeks[0] = indeks[--nr];
				}
				go_down(0);				
			}
		}
		else {
			while (nr > 0) {
				i = indeks[0];
				fprintf(output, "%s\n", first[i]);
				if (ferror(output)) {
					fprintf(stderr, "Blad!\n");
					return -1;
				}
				if (fscanf(files[i], "%s\n", first[i]) != 1) {
					if (ferror(files[i])) {
						fprintf(stderr, "Blad odczytu!\n");
						return -1;
					}
				   else indeks[0] = indeks[--nr];
				}
				go_down(0);				
			}
		}

		//sprzatanie
		for (i = stop-start; i>=0; i--) {
			delete[] first[i];
			fclose(files[i]); 
		}
		delete[] indeks;
		delete[] first;
		delete[] files;

		// pliki tymczasowe
		if (tmp) {
			sprintf(fname, "rm %s_*", prefiks);
			system(fname);
			fprintf(stderr, "%s\n", fname);
		}
	}
	
	//sprztanie - jeli otworzone pliki - to zamkn
	if (fileno(input) != fileno(stdin)) fclose(input);
	if (fileno(output) != fileno(stdout)) fclose(output);
	delete[] last;

	return 0;
}
